// Version 3.0.0

//###########################################################################
//###########################################################################
//####                                                                   ####
//####                           Olivier Pécheux                         ####
//####                          Olivier@Pecheux.fr                       ####
//####                         (33) +6 69 77 82 58                       ####
//####         http://dansetrad.fr/Orgue_de_barbarie/PecheuxGraph        ####
//####                                                                   ####
//###########################################################################
//###########################################################################

// Ceci est un complément aux pages de définition et d'explications de la
// librairie PecheuxGraph. Ce programme est une démonstration décrite dans la
// page http://dansetrad.fr/Orgue_de_barbarie/PecheuxGraph/ScreenShoot.html

// Ce programme permet de dessiner un peu comme le paint de XP

#include <PecheuxGraph.h>

//###########################################################################
//###########################################################################
//####                               PAINT                               ####
//###########################################################################
//###########################################################################


//###########################################################################
//##                           Variables globales                          ##
//###########################################################################

enum valeurFonction : byte {CRAYON, PIPETTE, PINCEAU, GOMME, DROITE, CERCLE, 
                            RECTANGLE, DECALAGE, REMPLISSAGE, BOMBE, LETTRES};

word couleur_trace = BLACK; // Couleur des tracés (crayon, pinceau, cercle...)
word couleur_fond = WHITE; // Couleur du fond (gomme, effacement d'écran)
boolean dessineEfface = true; // Etat du dernier choix  true=forme / false=fond


//###########################################################################
//##                              RadioInverse                             ##
//###########################################################################
// Boutons radio qui inversent la couleur quand ils sont sélectionnés. Comme
// il faut redéfinir draw pour une dizaine de boutons, c'est plus simple de
// faire une classe fille
class RadioInverse: public RadioZone
{
  public:
    RadioInverse(int x1, int y1, int x2, int y2, byte valeur);
    virtual void drawButton();
};

RadioInverse::RadioInverse(int x1, int y1, int x2, int y2, byte valeur): RadioZone(x1, y1, x2, y2, valeur) {}

void RadioInverse::drawButton()
{
  setDrawInverse(INVERSE); // Pour dessiner le curseur
  fillRect(demiX1 << 1, demiY1 << 1, demiX2 << 1, demiY2 << 1);
  setDrawInverse(NORMAL);
}


//###########################################################################
//##                          Choix de la couleur                          ##
//###########################################################################
// C'est le contrôle lié aux couleurs en bas. Il n'y a qu'un seul contrôle où
// il faut redéfinir la méthode onSelect, il vaut mieux utiliser une fonction
// externe
boolean onSelectCouleur()
{
  int x = getTouchX(); int y = getTouchY(); // Récupère les coordonnés visées
  croixCurseur(x, y); // Dessiner la croix curseur
  x = ((x + 6) & 0xFFF0) + 2;; y = (y > 219 ? 228 : 212); // Arrondir les valeurs au milieu des cases
  if (dessineEfface)
    fillRect(67, 212, 77, 222, couleur_trace = getPoint(x, y)); // Mémorise la couleur du dessin
  else
    fillRect(74, 219, 84, 229, couleur_fond = getPoint(x, y)); // Mémorise la couleur du fond
  return true;
}


//###########################################################################
//##                             Forme ou fond?                            ##
//###########################################################################
// C'est le double carré à gauche de la palette. Une seule instance, fonction
// externe pour le onSelect. C'est surtout parce qu'on n'a pas deux zones
// rectangulaires pour cliquer, mais deux triangles l'un forme="plus vers le
// point en haut à gauche" et fond="plus en bas à droite"

void onSelectFormeFond()
{
  int x = getTouchX(); int y = getTouchY(); // Récupère les coordonnés visées
  croixCurseur(x, y); // Dessiner la croix curseur
  if (x + y > 75 + 220) // La droite entre les 2 zones est y=x-75-220 Elle partage
    // le plan en deux un côté négatif, un côté positif
  { // Demande de sélection couleur de la gomme
    dessineEfface = false;
    fillRect(72, 217, 85, 230, WHITE);
    fillRect(73, 218, 85, 230, LIGHT_GREY);
    fillRect(74, 219, 84, 229, couleur_fond);
  }
  else
  { // Demande de sélection couleur du dessin
    dessineEfface = true;
    fillRect(66, 211, 79, 224, GREY);
    fillRect(66, 211, 78, 223, LIGHT_GREY);
    fillRect(67, 212, 77, 222, couleur_trace);
  }
}


//###########################################################################
//##                              Save et Load                             ##
//###########################################################################
// 2 boutons save et load, fonctions externes
void onSelectSave()
{
  String nom = clavier40(79, 10, F("Sauve sous?")); // Nom du fichier de sauvegarde
  bmpSave(nom, 68, 2, 317, 196);
  unselectRadio();
}


void onSelectLoad()
{
  String nom = clavier40(79, 10, F("Lire?")); // Nom du fichier de lecture
  bmpDraw(nom, 68, 2);
  unselectRadio();
}


//###########################################################################
//##                                 Efface                                ##
//###########################################################################
void onSelectEfface()
{
  fillRect(68, 2, 317, 196, couleur_fond);
  unselectRadio();
}


//###########################################################################
//##                               Decalages                               ##
//###########################################################################
void decalepoint(int dex, int dey, int versx, int versy)
{
  int couleur = getPoint(dex, dey);
  if (couleur != couleur_fond)
  {
    point(versx, versy, couleur); // dessine le nouveau point
    point(dex, dey, couleur_fond); // efface l'ancien
  }
}
void decale(int x1, int y1, int x2, int y2, int nouveauX, int nouveauY)
{
  if (x1 > x2) { x1 ^= x2; x2 ^= x1; x1 ^= x2; } // Echange éventuel pour avoir x2 à droite
  if (y1 > y2) { y1 ^= y2; y2 ^= y1; y1 ^= y2; } // Echange éventuel pour avoir y2 en bas
  int largeur = x2 - x1, hauteur = y2 - y1; // Taille de la zone à déplacer
  if (nouveauX < x1) // La nouvelle image est plus sur la gauche, on déplace les points en commençant par la gauche
    for (int x = 0; x < largeur; x++) for (int y = 0; y < hauteur; y++) // Pour chaque point
      decalepoint(x1 + x, y1 + y, nouveauX + x, nouveauY + y);
  else if (nouveauX > x1) // La nouvelle image est plus sur la droite, on déplace les points en commençant par la droite
    for (int x = largeur; x >= 0; x--) for (int y = 0; y < hauteur; y++) // Pour chaque point
      decalepoint(x1 + x, y1 + y, nouveauX + x, nouveauY + y);
  else if (nouveauY < y1) // La nouvelle image est parfaitement au dessus, on déplace les points en commençant par le haut
    for (int y = 0; y < hauteur; y++) for (int x = largeur; x >= 0; x--) // Pour chaque point
      decalepoint(x1 + x, y1 + y, nouveauX + x, nouveauY + y);
  else if (nouveauY > y1) // La nouvelle image est parfaitement en dessous, on déplace les points en commençant par le bas
    for (int y = hauteur; y >= 0; y--) for (int x = largeur; x >= 0; x--) // Pour chaque point
      decalepoint(x1 + x, y1 + y, nouveauX + x, nouveauY + y);
  // else l'image est au même endroit
}


//###########################################################################
//##                                Dessin                                 ##
//###########################################################################
int x2, y2, // Valeur actuelles
    x1, y1; // Valeur une boucle plus tôt
String s; // Pour les tracés de cercles
boolean onSelectDessin() // x y: valeur à l'entrée dans la fonction
{
  int x = getTouchX(); int y = getTouchY(); // Récupère les coordonnés visées
  x2 = x; y2 = y; // Au départ tous les points sont confondus; x1 et y1 sont initialisés plus loin
  setLimites(68, 2, 317, 196); // A la zone de dessin
  // Initialisations pour certains tracés
  switch (getRadioValeur())
  {
    case CRAYON:
      line(x, y, x, y, couleur_trace); // Initialise pour lineTo
      break;
    case DROITE:
    case CERCLE:
      setDrawInverse(INVERSE); // On va faire un curseur droite
      break;
    case RECTANGLE:
    case DECALAGE:
      setDrawInverse(INVERSE); // On va faire un curseur droite
      rect(x, y, x, y); // Dessine car on va l'effacer
      break;
  }
  do // On boucle sans passer par le gestionnaire, trop lent
  {
    x1 = x2, y1 = y2; //Sauvegarde de la valeur du coup d'avant
    x2 = getTouchX(); y2 = getTouchY(); // Utile pour la boucle
    switch (getRadioValeur())
    {
      case CRAYON:
        lineTo(x2, y2);
        break;
      case PIPETTE:
        setLimites(0, 0, 319, 239); // Pour pouvoir remplir la palette du choix
        if (dessineEfface) fill(69, 213, couleur_trace = getPoint(x, y)); // On change la couleur du dessin
        else fill(82, 227, couleur_fond = getPoint(x, y)); // On change la couleur du fond
        break;
      case PINCEAU:
      case GOMME:
        fillCircle(x2, y2, 3, getRadioValeur() == 2 ? couleur_trace : couleur_fond); // Couleur fond ou forme
        break;
      case DROITE:
      case CERCLE:
        line(x, y, x1, y1); // Efface l'ancienne
        line(x, y, x2, y2); // Dessine la nouvelle
        break;
      case RECTANGLE:
      case DECALAGE:
        rect(x, y, x1, y1); // Efface l'ancien
        rect(x, y, x2, y2); // Dessine le nouveau
        break;
      case REMPLISSAGE:
        fill(x2, y2, couleur_trace);
        break;
      case BOMBE:
        { // Accolades pour pouvoir utiliser des variables locales
          float angle = random(10000); angle = angle * PI / 5000;
          byte rayon = random(15);
          point(x2 + rayon * cos(angle), y2 + rayon * sin(angle), couleur_trace);
          break;
        }
      case LETTRES:
        s = clavier40(79, 10, F("Chaine?"));
        setTextSize(clavier40(79, 10, F("Taille?")).toInt());
        setTextColor(couleur_trace);
        setTextCursor(x2, y2);
        text(s);
    }

  } while (isTouch(100)); // Fini si on lâche le stylet
  switch (getRadioValeur()) // Onvient de lâcher le stylet
  {
    case DROITE:
      setDrawInverse(NORMAL); // Pour dessiner
      line(x, y, x2, y2, couleur_trace); // Tracé
      break;
    case CERCLE:
      line(x, y, x2, y2); // Effacel'ancienne à cause des cercles
      setDrawInverse(NORMAL); // Pour dessiner
      circle((x + x2) / 2, (y + y2) / 2, sqrt(pow(x - x2, 2) + pow(y - y2, 2)) / 2, couleur_trace); // Tracé
      break;
    case RECTANGLE:
      setDrawInverse(NORMAL); // Pour dessiner
      rect(x, y, x2, y2, couleur_trace); // Tracé du rectangle
      break;
    case DECALAGE:
      while (!isTouch()); // On attend un appui ailleurs
      x1=x; y1=y; // Mémorise le premier rectangle: (x1, y1, x2, y2)
      int x3 = getTouchX(), y3 = getTouchY(); // Récupère le point de départ
      x=x3; y=y3;
      rect(x1, y1, x2, y2); // Dessine la nouvelle position (efface le dernier rect car il y a supperposition
      do // On boucle sans passer par le gestionnaire, trop lent tant qu'on a pas la nouvelle position
      {
        rect(x-x3+x1, y-y3+y1, x-x3+x2, y-y3+y2); // Efface l'ancien
        x = getTouchX(); y = getTouchY(); // Nouveau point
        rect(x-x3+x1, y-y3+y1, x-x3+x2, y-y3+y2); // Dessine le nouveau
      } while (isTouch(100)); // Fini si on lâche le stylet
      rect(x1, y1, x2, y2); // Efface le premier rectangle de sélection
      rect(x-x3+x1, y-y3+y1, x-x3+x2, y-y3+y2); // Efface le deuxième rectangle de sélection
      setDrawInverse(NORMAL); // Pour dessiner
      decale(x1, y1, x2, y2, x-x3+x1, y-y3+y1); // Déclage du rectangle
  }
  setLimites(0, 0, 319, 239);
}



//###########################################################################
//##                                  Main                                 ##
//###########################################################################

PushZone *zone; // Variable temporaire pour certains contrôles
void setup()
{
  setGraphMode(PAYSAGE); // Initialisation générale
  setTextKeep(CUT); // Pour ne pas effacer l'écran si on met du setTexte qui dépasse

  // Fond d'écran
  clrscr(65534); // Couleur jaunâtre
  if (bpxDraw(F("DEMOS/PAINT1.BPX"), 9, 3) + // Dessin des symboles à gauche
      bpxDraw(F("DEMOS/PAINT2.BPX"), 12, 205)) // Dessin des couleurs en bas
  {
    text(F("\n Je ne touve pas les fichiers\n DEMOS/PAINT*.BPX"));
    while (1); // Arrêt du programme
  }
  rect(67, 1, 318, 197, BLACK); // Zone du dessin
  fillRect(68, 2, 317, 196, WHITE);

  zone = new PushZone(90, 203, 315, 236); // Palette de couleurs
  zone->onSelectFunction = &onSelectCouleur; // Quand on clique dessus

  zone = new PushZone(60, 203, 90, 236); // Choix couleur forme ou fond
  zone->onSelectFunction = &onSelectFormeFond; // Quand on clique dessus

  zone = new PushZone(5, 46, 30, 72); // Load
  zone->onSelectFunction = &onSelectLoad; // Quand on clique dessus

  zone = new PushZone(32, 46, 58, 72); // Save
  zone->onSelectFunction = &onSelectSave; // Quand on clique dessus

  zone = new PushZone(32, 182, 58, 208); // Efface
  zone->onSelectFunction = &onSelectEfface; // Quand on clique dessus

  zone = new PushZone(68, 2, 317, 196); // zone de dessin
  zone->onSelectFunction = &onSelectDessin; // Quand on clique dessus

  new RadioInverse(5,  74,  30, 99,  CRAYON);
  new RadioInverse(5,  101, 30, 126, PIPETTE);
  new RadioInverse(5,  128, 30, 154, PINCEAU);
  new RadioInverse(5,  156, 30, 180, GOMME);
  new RadioInverse(5,  182, 30, 208, DROITE);
  new RadioInverse(32, 210, 58, 233, CERCLE);
  new RadioInverse(5,  210, 30, 233, RECTANGLE);
  new RadioInverse(32, 74,  58, 99,  DECALAGE);
  new RadioInverse(32, 101, 58, 126, REMPLISSAGE);
  new RadioInverse(32, 128, 58, 154, BOMBE);
  new RadioInverse(32, 156, 58, 180, LETTRES);

  // drawControles(); les boutons sont des boutons à inversion déja dessinés. Ne pas appeler drawControle()
}


void loop()
{
  scanEvent(); // Appel récurent du gestionnaire touchpad
}
