// Version 3.0.0

//###########################################################################
//###########################################################################
//####                                                                   ####
//####                           Olivier Pécheux                         ####
//####                          Olivier@Pecheux.fr                       ####
//####                         (33) +6 69 77 82 58                       ####
//####         http://dansetrad.fr/Orgue_de_barbarie/PecheuxGraph        ####
//####                                                                   ####
//###########################################################################
//###########################################################################

// Ceci est un complément aux pages de définition et d'explications de la
// librairie PecheuxGraph. Ce programme illustre l'utilisation de la classe
// Timer décrites dans la page
// http://dansetrad.fr/Orgue_de_barbarie/PecheuxGraph/Definition/TimerExemple.html

// L'intérieur d'une maison est dessinée. Deux boutons de type PushZone
// permettent d'allumer la lumière pendant 6s

// En principe scanEvent() est seul dans loop(). Ici c'est une exception, on
// va profiter de loop pour afficher l'état du temporisateur.

#include <PecheuxGraph.h>

// Définition des objets
PushZone bouton1(186, 124, 212, 152); // Définit une zone cliquable, notre bouton de droite
PushZone bouton2(8, 124, 34, 152); // Définit une zone cliquable, notre bouton de gauche
Timer minuterie(6000); // Minuterie principale de 6 secondes


void aiguilleDuChronometre(void) // Le chronomètre au dessus de la porte
{
  if (minuterie.actif) // Si on compte le temps
  {
    float angle = ((millis() - minuterie.depart) * PI) / (minuterie.duree_ms >> 1); // angle que fait la ligne à tracer
    line(264, 26, 264 + 9 * sin(angle), 26 - 9 * cos(angle), RED); // Augmente le camembert rouge
  }
}


// Quand on appuie sur un bouton, on déclenche la minuterie, on dessine le chronomètre,
// et on allume la lampe
void debutMinuterie()
{
  // Démarre la temporisation
  minuterie.start();
  // Dessin de la minuterie (pour faire patienter)
  fillCircle(264, 26, 10, GREY); // Corps du chronomètre (efface le rouge si il en avait)
  circle(264, 26, 10, BLACK); // Monture
  fillCircle(255, 17, 2); fillCircle(264, 13, 2); fillCircle(273, 17, 2); // 3 boutons
  // On allume la lampe
  fill(160, 30, YELLOW);
}


// Quand le temps est écoulé, on efface le chronomètre et on éteint la lampe
void finMinuterie()
{
  fillRect(253, 11, 275, 36, GREY); // On efface le chronomètre
  fill(160, 30, BLACK); // On éteint la lampe
}


void setup()
{
  setGraphMode(PAYSAGE); // Initialisation

  // Consigne tout en bas
  setTextCursor(0, 180);
  text(F("\n Appuyez sur les boutons de\n  chaque côté de la fentêre"));

  maison(1, 1); // On dessine l'intérieur de la maison
  fillRect(253, 11, 275, 36, GREY); // Boite en plastique contenant le chronomètre

  // Mise en place des comportements des objets
  bouton1.onSelectFunction = &debutMinuterie; // Bouton de droite, c'est le début de la minuterie
  bouton2.onSelectFunction = &debutMinuterie; // Bouton de gauche, c'est le début de la minuterie
  minuterie.onTimeFunction = &finMinuterie; // En fin de temporisation tout s'arrête

}

void loop()
{
  scanEvent(); // Lance le gestionnaire
  aiguilleDuChronometre(); // dessine le camembert du chronmètre
}
