// Version 3.0.0

//###########################################################################
//###########################################################################
//####                                                                   ####
//####                           Olivier Pécheux                         ####
//####                          Olivier@Pecheux.fr                       ####
//####                         (33) +6 69 77 82 58                       ####
//####         http://dansetrad.fr/Orgue_de_barbarie/PecheuxGraph        ####
//####                                                                   ####
//###########################################################################
//###########################################################################


// Ceci est un complément aux pages de définition et d'explications de la
// librairie PecheuxGraph. Ce programme illustre l'utilisation de la classe
// CheckZone décrites dans la page
// http://dansetrad.fr/Orgue_de_barbarie/PecheuxGraph/Definition/CheckZoneExemple.html

// Mise em place de 3 boutons de type CheckZone (zone à cocher, pas de dessin
// dans le bouton). Quand on appuie sur un bouton, un disque blanc montre son
// activité

// Version 1:
// - Les boutons ont une définition statique
// - Utilise onSelectFunction pour exécuter l'action à faire
// - Une action différente pour chaque bouton

#include <PecheuxGraph.h> // Appel de la bibliothèque 


// Déclaration statique des instances, doit se faire en dehors du setup
// Dans ce cas les boutons sont accessibles partout
CheckZone boutonHaut(200,40,240,80); // Bouton en haut
CheckZone boutonCentre(200,100,240,140); // Bouton au centre
CheckZone boutonBas(200,160,240,200); // Bouton en bas


// Déclaration de 3 fonctions, une par bouton, qui feront ce qu'il faut
// faire à chaque changement
// Ces fonctions ne doivent pas avoir de paramètres et ne doit rien retourner
// Les noms peuvent être quelconques
void onSelectActionHaut() // Action à faire quand on appuie sur le bouton du haut
{
  if (boutonHaut.isSelected()) fillCircle(100,60,20,WHITE);
  else fillCircle(100,60,20,BLACK);
}


void onSelectActionCentre() // Action à faire quand on appuie sur le bouton du centre
{
  if (boutonCentre.isSelected()) fillCircle(100,120,20,WHITE);
  else fillCircle(100,120,20,BLACK);
}


void onSelectActionBas() // Action à faire quand on appuie sur le bouton du bas
{
  if (boutonBas.isSelected()) fillCircle(100,180,20,WHITE);
  else fillCircle(100,180,20,BLACK);
}


void setup()
{
  setGraphMode(PAYSAGE); // Initialisation de la carte
  rect(200,40,240,80,GREEN); // Carré du haut
  rect(200,100,240,140,GREEN); // Carré au centre de l'écran
  rect(200,160,240,200,GREEN); // Carré du bas
  text(F("Appuyez sur un carré"));
  
  boutonHaut.onSelectFunction=&onSelectActionHaut; // Fonction appelée si on active le bouton du haut de l'écran
  boutonCentre.onSelectFunction=&onSelectActionCentre; // Fonction appelée si on active le bouton au milieu de l'écran
  boutonBas.onSelectFunction=&onSelectActionBas; // Fonction appelée si on active le bouton en bas de l'écran
  boutonHaut.onUnselectFunction=&onSelectActionHaut; // Fonction appelée si on désactive le bouton du haut de l'écran
  boutonCentre.onUnselectFunction=&onSelectActionCentre; // Fonction appelée si on désactive le bouton au milieu de l'écran
  boutonBas.onUnselectFunction=&onSelectActionBas; // Fonction appelée si on désactive le bouton en bas de l'écran
}


void loop() 
{
  scanEvent(); // Gestion des boutons et des horloges, le plus souvent seul dans le loop
}
