// Version 3.0.0

//###########################################################################
//###########################################################################
//####                                                                   ####
//####                           Olivier Pécheux                         ####
//####                          Olivier@Pecheux.fr                       ####
//####                         (33) +6 69 77 82 58                       ####
//####         http://dansetrad.fr/Orgue_de_barbarie/PecheuxGraph        ####
//####                                                                   ####
//###########################################################################
//###########################################################################

// Ceci est un complément aux pages de définition et d'explications de la
// librairie PecheuxGraph. Ce programme illustre l'utilisation de la classe
// RadioZone décrites dans la page
// http://dansetrad.fr/Orgue_de_barbarie/PecheuxGraph/Definition/RadioZoneExemple.html

// L'intérieur d'une maison est dessinée. Trois boutons de type Radio
// permettent de fermer le volet gauche, trois autres permetent de fermer le
// volet droit

#include <PecheuxGraph.h>

// Equivalece des constantes
const byte MONTE = 0;
const byte STOP = 2;
const byte DESCEND = 1;

// Définition des contrôles dynamiques, peut se faire avant ou après setup
// On peut panacher, cela n'a pas d'importance. Normalement, on ne le fait pas.
// Si c'est fait ici, c'est à titre d'exemple.

// Bouton de droite, groupe 0 (pas nécessaire de le préciser, c'est par défaut)
RadioZone *monteDroite = new RadioZone(184, 115, 210, 127, MONTE); // Définit une zone cliquable, notre bouton monte de droite
// Notez que monteDroite est un nom qui ne nous servira pas, mais on en a besoin pour la déclaration

// Boutons de gauche, groupe 1, il faut le dire
RadioZone *stopGauche = new RadioZone(7, 129, 35, 141, STOP, GROUPE_1); // Définit une zone cliquable, notre bouton stop de gauche
RadioZone *descendGauche = new RadioZone(7, 143, 35, 155, DESCEND, GROUPE_1); // Définit une zone cliquable, notre bouton descend de gauche

// On va garder l'horloge satique. Le panachage est complet.
Clock horloge(100); // Le volet se déplace toutes les 100ms



// Variables et constantes pour les volets roulants
const word BLEU_CIEL = 0x755E; // Couleur du ciel
const byte HAUT = 55; // Le haut de la fenêtre
const byte BAS = 155; // le bas de la fenêtre
byte positionDroite = HAUT - 1; // Position du volet droit (plus haut que HAUT, donc complètement rentré)
byte positionGauche = HAUT - 1; // Position du volet gauche (plus haut que HAUT, donc complètement rentré)

void onTimeVolets(void) // Fonction appelée tous les 100 ms pour redessiner les volets
{
  //------------ Volet de droite ----------------
  switch (getRadioValeur()) // Gestion du mouvement
  {
    case MONTE:
      if (positionDroite >= HAUT) positionDroite--; // Monte le volet d'un pixel si il n'est pas en haut
      else // on vient d'arriver en haut
      {
        // Désactivation du bouton haut, choisir une des deux possibilités (une seule sans commentaires)
        // getRadioPointeur()->unselect(); // Désactive le bouton activé, c'est le bouton haut
        unselectRadio(); // Désactive tous les boutons
        if (positionGauche == HAUT - 1) // si l'autre volet est aussi en haut
          fill(160, 30, BLACK); // On éteint la lampe
      }
      break;
    case DESCEND:
      if (positionDroite < BAS) positionDroite++; // Descend le volet d'un pixel si il n'est pas en haut
      else // on vient d'arriver en bas
      {
        // Désactivation du bouton haut, choisir une des deux possibilités (une seule sans commentaires)
        // getRadioPointeur()->unselect(); // Désactive le bouton activé
        unselectRadio(); // Désactive tous les boutons
        if (positionGauche == BAS) // si l'autre volet est aussi en bas
          fill(160, 30, YELLOW); // On allume la lampe
      }
  }
  // Dessin du volet
  if ((getRadioValeur() == MONTE) || (getRadioValeur() == DESCEND)) // Sinon cela clignotte aussi à l'arrêt
  {
    // Dessine toutes les lattes en marron
    fillRect(114, HAUT, 170, positionDroite, DARK_RED);
    // dessine une latte sur 16 en noir
    for (byte latte = 0; latte <= positionDroite - HAUT; latte += 16) hLine(114, 170, positionDroite - latte, BLACK); // Une latte sur 16 est noire
    // Efface la latte en dessous (au cas ou on remonte le volet)
    if (positionDroite != BAS) hLine(114, 170, positionDroite + 1, BLEU_CIEL); // Seulement si on n'est pas tout en bas, sinon on dessinerait dans le cadre
  }

  //------------ Volet de gauche ----------------
  switch (getRadioValeur(GROUPE_1)) // Gestion du mouvement
  {
    case MONTE: 
      if (positionGauche >= HAUT) positionGauche--; // Monte le volet d'un pixel si il n'est pas en haut
      else // on vient d'arriver en haut
      {
        // Désactivation du bouton haut, choisir une des deux possibilités (une seule sans commentaires)
        // getRadioPointeur()->unselect(); // Désactive le bouton activé, c'est le bouton haut
        unselectRadio(GROUPE_1); // Désactive tous les boutons
        if (positionDroite == HAUT - 1) // si l'autre volet est aussi en haut
          fill(160, 30, BLACK); // On éteint la lampe
      }
      break;
    case DESCEND: 
      if (positionGauche < BAS) positionGauche++; // Descend le volet d'un pixel si il n'est pas en haut
      else // on vient d'arriver en bas
      {
        // Désactivation du bouton haut, choisir une des deux possibilités (une seule sans commentaires)
        // getRadioPointeur()->unselect(); // Désactive le bouton activé
        unselectRadio(GROUPE_1); // Désactive tous les boutons
        if (positionDroite == BAS) // si l'autre volet est aussi en bas
          fill(160, 30, YELLOW); // On allume la lampe
      }
  }
  // Dessin du volet
  if ((getRadioValeur(GROUPE_1) == MONTE) || (getRadioValeur(GROUPE_1) == DESCEND)) // Sinon cela clignotte aussi à l'arrêt
  {
    // Dessine toutes les lattes en marron
    fillRect(50, HAUT, 107, positionGauche, DARK_RED);
    // dessine une latte sur 16 en noir
    for (byte latte = 0; latte <= positionGauche - HAUT; latte += 16) hLine(50, 107, positionGauche - latte, BLACK); // Une latte sur 16 est noire
    // Efface la latte en dessous (au cas ou on remonte le volet)
    if (positionGauche != BAS) hLine(50, 107, positionGauche + 1, BLEU_CIEL); // Seulement si on n'est pas tout en bas, sinon on dessinerait dans le cadre
  }
}

void setup()
{
  setGraphMode(PAYSAGE); // Initialisation

  // Définition des contrôles dynamiques, après setup
  // Comme on n'a pas besoin de changer un comportement, on n'a pas besoin de garder une trace
  // de l'objet. On se moque de ce que retourne new

  // Boutons de droite, groupe 0 (pas nécessaire de le préciser, c'est par défaut)
  new RadioZone(184, 129, 210, 141, STOP); // Définit une zone cliquable, notre bouton stop de droite
  new RadioZone(184, 143, 210, 155, DESCEND); // Définit une zone cliquable, notre bouton descend de droite

  // Boutons de gauche, groupe 1, il faut le dire
  new RadioZone(7, 115, 35, 127, MONTE, GROUPE_1); // Définit une zone cliquable, notre bouton monte de gauche

  horloge.onTimeFunction = &onTimeVolets; // Action à faire quand l'horloge donne son tip

  // Consigne
  setTextCursor(10, 200);
  text(F("Appuyez sur les boutons,\n de chaque côté de la porte"));

  maison(3, 3); // avec 3 boutons de chaque côté
  fillRect(50, 55, 107, 155, BLEU_CIEL); // Vitre de gauche, ciel uni
  fillRect(114, 55, 170, 155); // Vitre de droite, ciel uni
}


void loop()
{
  scanEvent(); // Gestion des boutons et des horloges, le plus souvent seul dans le loop
}
